#include "DXBuffer.h"

#include <iostream>
#ifdef WIN32
#    include <d3d11.h>
#    include <d3d12.h>

#    include "d3dx12.h"
#endif

using namespace Framework;

// Inhalt der DXBuffer Klasse

// Konstruktor
//  bind: Der verwendungszweck des Buffers. Beispiel: D3D11_BIND_INDEX_BUFFER,
//  D3D11_BIND_VERTEX_BUFFER. eLn: Lnge eines einzelnen Elements in Bytes
DXBuffer::DXBuffer(int eLen)
    : ReferenceCounter()
{
    data = 0;
    changed = 0;
    len = 0;
    elLen = eLen;
}

// Destruktor
DXBuffer::~DXBuffer() {}

// Setzt den gendert flg, so das beim nchsten auruf von 'kopieren' die daten
// neu kopiert werden
void DXBuffer::setChanged()
{
    changed = 1;
}

// ndert die lnge des Buffers beim nchsten aufruf von 'kopieren'
//  ln: Die Lnge in Bytes
void DXBuffer::setLength(int len)
{
    this->len = len;
}

// Legt fest, was beim nchsten aufruf von 'kopieren' kopiert wird
//  data: Ein zeiger auf die Daten
void DXBuffer::setData(void* data)
{
    this->data = data;
    changed = 1;
}

// Gibt die Lnge eines Elementes in bytes zurck
int DXBuffer::getElementLength() const
{
    return elLen;
}

// Gibt die Anzahl der Elemente im Buffer zurck
int DXBuffer::getElementAnzahl() const
{
    return len / elLen;
}

#ifdef WIN32
// Inhalt der DX11Buffer Klasse

// Konstruktor
// eSize: Die Lnge eines Elementes in Bytes
DX11Buffer::DX11Buffer(int eSize,
    ID3D11Device* device,
    ID3D11DeviceContext* context,
    int bindFlags)
    : DXBuffer(eSize)
{
    buffer = 0;
    description = new D3D11_BUFFER_DESC();
    memset(description, 0, sizeof(description));
    description->Usage = D3D11_USAGE_DYNAMIC;
    description->CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
    description->BindFlags = bindFlags;

    this->device = device;
    this->context = context;
}

// Destruktor
DX11Buffer::~DX11Buffer()
{
    if (buffer) buffer->Release();
    delete description;
}

// Kopiert die Daten in den Buffer, fals sie sich verndert haben
//  zRObj: Das Objekt, mit dem die Grafikkarte angesprochen wird
void DX11Buffer::copieren(int byteCount)
{
    if (!len) return;
    if (byteCount < 0) byteCount = len;
    if (description->ByteWidth < (unsigned)len)
    {
        if (buffer) buffer->Release();
        buffer = 0;
    }
    if (!buffer)
    {
        description->ByteWidth = len;
        device->CreateBuffer(description, 0, &buffer);
        if (data) changed = 1;
    }
    if (changed)
    {
        D3D11_MAPPED_SUBRESOURCE map;
        if ((description->Usage | D3D11_USAGE_DYNAMIC) == description->Usage)
            context->Map(
                buffer, 0, D3D11_MAP::D3D11_MAP_WRITE_DISCARD, 0, &map);
        else
            context->Map(buffer, 0, D3D11_MAP::D3D11_MAP_WRITE, 0, &map);
        memcpy(map.pData, data, byteCount);
        context->Unmap(buffer, 0);
        changed = 0;
    }
}

// Gibt den Buffer zurck
ID3D11Buffer* DX11Buffer::zBuffer() const
{
    return buffer;
}

// Inhalt der DXStructuredBuffer Klasse

// Konstruktor
// eSize: Die Lnge eines Elementes in Bytes
DX11StructuredBuffer::DX11StructuredBuffer(
    int eSize, ID3D11Device* device, ID3D11DeviceContext* context)
    : DX11Buffer(eSize,
        device,
        context,
        D3D11_BIND_UNORDERED_ACCESS | D3D11_BIND_SHADER_RESOURCE)
{
    description->MiscFlags = D3D11_RESOURCE_MISC_BUFFER_STRUCTURED;
    description->StructureByteStride = eSize;
    description->Usage = D3D11_USAGE_DEFAULT;
    view = 0;
}

// Destruktor
DX11StructuredBuffer::~DX11StructuredBuffer()
{
    if (view) view->Release();
}

// Kopiert die Daten in den Buffer, fals sie sich verndert haben
//  zRObj: Das Objekt, mit dem die Grafikkarte angesprochen wird
void DX11StructuredBuffer::copieren(int byteCount)
{
    ID3D11Buffer* old = buffer;
    DX11Buffer::copieren(byteCount);
    if (buffer != old)
    {
        if (view) view->Release();
        D3D11_SHADER_RESOURCE_VIEW_DESC desc = {};
        desc.ViewDimension = D3D11_SRV_DIMENSION_BUFFEREX;
        desc.BufferEx.FirstElement = 0;
        desc.Format = DXGI_FORMAT_UNKNOWN;
        desc.BufferEx.NumElements
            = description->ByteWidth / description->StructureByteStride;
        device->CreateShaderResourceView(buffer, &desc, &view);
    }
}

// Gibt die verwendtete Shader Resource View zurck
DX11StructuredBuffer::operator ID3D11ShaderResourceView*() const
{
    return view;
}

#endif